from ayoomoney import auth
import click


@click.group()
def main():
    pass


@main.command()
@click.argument("client_id")
@click.argument("redirect_url")
@click.option("--scope", default="", help="Список разрешений/scope, по умолчанию включены все разрешения")
def simple(client_id: str, redirect_url: str, scope: str):
    auth.simple.authorize(
        client_id,
        redirect_url,
        scope=scope.split(",") if scope else auth.simple.DEFAULT_SCOPE
    )


@main.command()
@click.argument("client_id")
@click.argument("redirect_url")
@click.option("--host", default="127.0.0.1")
@click.option("--port", default=80, help="Порт приложения")
@click.option("--scope", default="", help="Список разрешений/scope, по умолчанию включены все разрешения")
def auto(client_id: str, redirect_url: str, host: str, port: int, scope: str):
    auth.auto.authorize(
        client_id,
        redirect_url,
        host=host,
        port=port,
        scope=scope.split(",") if scope else auth.auto.DEFAULT_SCOPE
    )


if __name__ == '__main__':
    main()



from http.server import BaseHTTPRequestHandler
from contextlib import closing
import socketserver
import socket

from .base import Authorization, DEFAULT_SCOPE
from ayoomoney.errors import CreateTokenError

HOST = "127.0.0.1"
PORT = 80


def is_port_free(host, port) -> bool:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        if sock.connect_ex((host, port)) == 0:
            return False
        else:
            return True


class CodeHandler(BaseHTTPRequestHandler):
    client_id = None
    redirect_url = None
    auth_client: Authorization = None
    access_token: str | None = None

    def log_request(self, code: int | str = ..., size: int | str = ...) -> None:
        pass

    def do_GET(self):
        if "code=" not in self.path:
            self.send_response(400)
            self.end_headers()
            return

        code = self.path.split("code=")[-1]
        access_token = ""
        error = None

        try:
            access_token = self.auth_client.get_access_token(code)
        except CreateTokenError as e:
            error = e

        if error:
            print(error)
            body = f"""<div style="word-wrap: break-word;"><b>error:</b> {error}</div>"""
        else:
            body = f"""<div style="word-wrap: break-word;"><b>access_token:</b> {access_token}</div>"""

        print(f"{access_token=}")
        self.access_token = access_token
        self.send_response(200)
        self.send_header("charset", "windows-1251")
        self.end_headers()
        self.wfile.write(body.encode("windows-1251"))


def authorize(
        client_id: str,
        redirect_uri: str,
        scope: list[str] = DEFAULT_SCOPE,
        host: str = HOST, port: int = PORT,
        *_
) -> str | None:
    if not is_port_free(host, port):
        print(
            f"Порт: {port} занят другим приложением. "
            f"Попробуйте закрыть его или укажите другой порт командой: `--port N`\n"
        )
        print(
            "После изменения порта нужно изменить redirect_uri приложения, зайдите на страницу: "
            "https://yoomoney.ru/settings/oauth-services"
        )
        print("и в поле redirect_uri добавьте текущий порт: http://my.localhost:N")
        exit(1)

    parts = redirect_uri.split(":")
    if len(parts) > 2:
        _port = int(parts[-1])
        if _port != port:
            print(
                f"Порт в {redirect_uri=} не совпадает с текущим: {port}. "
                f"Укажите порт {_port} командой: `--port {_port}`"
            )
            exit(1)

    auth = Authorization(client_id, redirect_uri)
    url = auth.authorization_request(scope=scope)
    print("\n".join([
        "Перейдите по ссылке и подтвердите доступ для приложения:",
        url,
        "",
        "После подтверждения вы получите access_token, его можно скопировать с web-страницы или консоли.",
        f"Для отмены операции перейдите по адресу: http://{host}:{port}",
        ""
    ]))

    handler = CodeHandler
    handler.client_id = client_id
    handler.redirect_url = redirect_uri
    handler.auth_client = auth
    with socketserver.TCPServer((host, port), CodeHandler) as httpd:
        httpd.handle_request()

    return handler.access_token



from enum import Enum


class PaymentSource(str, Enum):
    BANK_CARD = "AC"
    YOOMONEY_WALLET = "PC"


class OperationDirection(str, Enum):
    IN = "in"
    OUT = "out"


class OperationStatus(str, Enum):
    SUCCESS = "success"
    REFUSED = "refused"
    IN_PROGRESS = "in_progress"


class OperationType(str, Enum):
    PAYMENT_SHOP = "payment-shop"
    OUTGOING_TRANSFER = "outgoing-transfer"
    DEPOSITION = "deposition"
    INCOMING_TRANSFER = "incoming-transfer"


class OperationHistoryParamType(str, Enum):
    DEPOSITION = "deposition"
    PAYMENT = "payment"


class RecipientType(str, Enum):
    ACCOUNT = "account"
    PHONE = "phone"
    EMAIL = "email"



from types import TracebackType
from typing import Self, Type

from httpx import AsyncClient, Response, Client

from ayoomoney.types import (
    AccountInfo,
    OperationDetails,
    OperationStatus,
    OperationHistory,
    PaymentSource,
    PaymentForm,
    OperationHistoryParams
)
import ayoomoney


class _BaseWallet:
    BASE_URL = "https://yoomoney.ru"

    def __init__(self, access_token: str, headers: dict | None = None):
        if headers is None:
            headers = {}

        self._headers = {
            "Authorization": f"Bearer {access_token}",
            **headers
        }

    def _process_account_info(self, response: Response) -> AccountInfo | None:
        if not response.is_success:
            if response.status_code == 401:
                raise ayoomoney.errors.InvalidTokenError
            return

        return AccountInfo.model_validate_json(response.content)

    def _process_get_operation_details(self, response: Response) -> OperationDetails | None:
        if not response.is_success:
            if response.status_code == 401:
                raise ayoomoney.errors.InvalidTokenError
            return

        return OperationDetails.model_validate_json(response.content)

    def _process_get_operation_history(self, response: Response) -> OperationHistory | None:
        if not response.is_success:
            if response.status_code == 401:
                raise ayoomoney.errors.InvalidTokenError
            return

        history = OperationHistory.model_validate_json(response.content)
        return history

    def _process_create_payment_form(self, unique_label: str, response: Response) -> PaymentForm:
        if response.status_code == 401:
            raise ayoomoney.errors.InvalidTokenError

        return PaymentForm(
            link_for_customer=str(response.url),
            payment_label=unique_label
        )

    def _process_check_payment_on_successful(self, history: OperationHistory) -> bool:
        if history is None or len(history.operations) <= 0:
            return False

        operation = history.operations[0]
        return operation.status == OperationStatus.SUCCESS


class YooMoneyWallet(_BaseWallet):
    def __init__(self, access_token: str, headers: dict | None = None):
        super().__init__(access_token, headers)
        self.client = Client(
            base_url=self.BASE_URL,
            headers=self._headers
        )

    def __enter__(self) -> Self:
        return self

    def __exit__(
        self,
        exc_type: Type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        self.close()

    def close(self):
        self.client.close()

    def account_info(self) -> AccountInfo | None:
        url = "/api/account-info"
        response = self.client.post(url)
        return self._process_account_info(response)

    def get_operation_details(self, operation_id: str) -> OperationDetails | None:
        url = "/api/operation-details"
        response = self.client.post(url, data={"operation_id": operation_id})
        return self._process_get_operation_details(response)

    def get_operation_history(self, params: OperationHistoryParams | None = None) -> OperationHistory | None:
        if params is None:
            params = OperationHistoryParams()

        url = "/api/operation-history"
        params = params.model_dump(exclude_none=True, by_alias=True)
        response = self.client.post(url, data=params)
        return self._process_get_operation_history(response)

    def create_payment_form(
            self,
            amount_rub: int,
            unique_label: str,
            success_redirect_url: str | None = None,
            payment_source: PaymentSource = PaymentSource.BANK_CARD
    ) -> PaymentForm:
        account_info = self.account_info()
        url = "/quickpay/confirm.xml"
        params = {
            "receiver": account_info.account,
            "quickpay-form": "button",
            "paymentType": payment_source,
            "sum": amount_rub,
            "successURL": success_redirect_url,
            "label": unique_label
        }
        params = {k: v for k, v in params.items() if v}
        response = self.client.post(url, params=params)
        return self._process_create_payment_form(unique_label, response)

    def check_payment_on_successful(self, label: str) -> bool:
        params = OperationHistoryParams(label=label)
        history = self.get_operation_history(params)
        return self._process_check_payment_on_successful(history)

    def revoke_token(self) -> bool:
        url = "/api/revoke"
        response = self.client.post(url)
        return response.is_success


class YooMoneyWalletAsync(_BaseWallet):
    def __init__(self, access_token: str, headers: dict | None = None):
        super().__init__(access_token, headers)
        self.client = AsyncClient(
            base_url=self.BASE_URL,
            headers=self._headers
        )

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: TracebackType | None = None,
    ) -> None:
        await self.close()

    async def close(self):
        await self.client.aclose()

    async def account_info(self) -> AccountInfo | None:
        url = "/api/account-info"
        response = await self.client.post(url)
        return self._process_account_info(response)

    async def get_operation_details(self, operation_id: str) -> OperationDetails | None:
        url = "/api/operation-details"
        response = await self.client.post(url, data={"operation_id": operation_id})
        return self._process_get_operation_details(response)

    async def get_operation_history(self, params: OperationHistoryParams | None = None) -> OperationHistory | None:
        if params is None:
            params = OperationHistoryParams()

        url = "/api/operation-history"
        params = params.model_dump(exclude_none=True, by_alias=True)
        response = await self.client.post(url, data=params)
        return self._process_get_operation_history(response)

    async def create_payment_form(
            self,
            amount_rub: int,
            unique_label: str,
            success_redirect_url: str | None = None,
            payment_source: PaymentSource = PaymentSource.BANK_CARD
    ) -> PaymentForm:
        account_info = await self.account_info()
        url = "/quickpay/confirm.xml"
        params = {
            "receiver": account_info.account,
            "quickpay-form": "button",
            "paymentType": payment_source,
            "sum": amount_rub,
            "successURL": success_redirect_url,
            "label": unique_label
        }
        params = {k: v for k, v in params.items() if v}
        response = await self.client.post(url, params=params)
        return self._process_create_payment_form(unique_label, response)

    async def check_payment_on_successful(self, label: str) -> bool:
        params = OperationHistoryParams(label=label)
        history = await self.get_operation_history(params)
        return self._process_check_payment_on_successful(history)

    async def revoke_token(self) -> bool:
        url = "/api/revoke"
        response = await self.client.post(url)
        return response.is_success



from datetime import datetime

from pydantic import BaseModel, Field

from .enums import OperationDirection, OperationStatus, OperationType, RecipientType


class OperationDetails(BaseModel):
    """
    Детальная информация об операции из истории
    https://yoomoney.ru/docs/wallet/user-account/operation-details
    """
    amount: float = Field(...)
    amount_due: float | None = Field(None)
    error: str | None = Field(None)
    operation_id: str = Field(...)
    direction: OperationDirection = Field(...)
    status: OperationStatus = Field(...)
    pattern_id: str | None = Field(None)
    fee: float | None = Field(None)
    title: str = Field(...)
    sender: int | None = Field(None)
    recipient: str | None = Field(None)
    recipient_type: RecipientType | None = Field(None)
    message: str | None = Field(None)
    comment: str | None = Field(None)
    codepro: bool | None = Field(None)
    label: str | None = Field(None)
    details: str | None = Field(None)
    digital_goods: dict | None = Field(None)
    operation_type: OperationType = Field(..., alias="type")
    execution_datetime: datetime = Field(..., alias="datetime")



class InvalidTokenError(Exception):
    def __init__(self):
        message = "Указанный токен не существует, либо отозван."
        super().__init__(message)


class CreateTokenError(Exception):
    def __init__(self, error: str):
        info = {
            "invalid_grant": "Обязательные параметры запроса отсутствуют или имеют некорректные или недопустимые значения.",
            "unauthorized_client": "Неверное значение параметра client_id или client_secret, либо приложение не имеет права запрашивать авторизацию (например, ЮMoney заблокировали его client_id).",
            "invalid_request": "В выдаче access_token отказано. ЮMoney не выдавали временный токен, токен просрочен, или по этому временному токену уже выдан access_token (повторный запрос токена авторизации с тем же временным токеном).",
        }
        message = "\n".join([
            f"Не удалось создать токен: {error=}",
            f"Описание ошибки: {info.get(error, 'Неизвестная ошибка')}"
        ])
        super().__init__(message)



from .enums import (
    OperationHistoryParamType,
    OperationDirection,
    OperationStatus,
    PaymentSource,
    OperationType,
    RecipientType,
)
from .operation_details import OperationDetails
from .operation_history import Operation, OperationHistory, OperationHistoryParams
from .account_info import AccountInfo
from .payment import PaymentForm




from datetime import datetime

from pydantic import BaseModel, Field, conint, field_serializer

from .enums import OperationDirection, OperationStatus, OperationType, OperationHistoryParamType
from .operation_details import OperationDetails
from .utils import convert_datetime_to_iso_8601


class Operation(BaseModel):
    """
    Описание платежной операции
    https://yoomoney.ru/docs/wallet/user-account/operation-history#response-operation
    """
    operation_id: str = Field(...)
    status: OperationStatus = Field(...)
    execution_datetime: datetime = Field(..., alias="datetime")
    title: str = Field(...)
    pattern_id: str | None = Field(None)
    direction: OperationDirection = Field(...)
    amount: float = Field(...)
    label: str | None = Field(None)
    operation_type: OperationType = Field(..., alias="type")


class OperationHistory(BaseModel):
    error: str | None = Field(None)
    next_record: int | None = Field(None)
    operations: list[Operation | OperationDetails] = Field(...)


class OperationHistoryParams(BaseModel):
    operation_type: list[OperationHistoryParamType] | None = Field(default=None, serialization_alias="type")
    label: str | None = Field(default=None)
    from_datetime: datetime | None = Field(default=None, serialization_alias="from")
    till_datetime: datetime | None = Field(default=None, serialization_alias="till")
    start_record: int | None = Field(default=None)
    records: conint(ge=1, le=100) | None = Field(default=None)
    details: bool | None = Field(default=None)

    @field_serializer("from_datetime", "till_datetime")
    def convert_datetime_to_iso8601(self, v: datetime, _info) -> str:
        return convert_datetime_to_iso_8601(v)

    @field_serializer("operation_type")
    def convert_types(self, v: list[str], _info) -> str:
        return " ".join(v)

    @field_serializer("start_record")
    def convert_to_string(self, v: int, _info) -> str:
        return str(v)
